/*
 * com/ngon/gfx/PixelDrawer.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.gfx;

/*
 * class PixelDrawer
 */
public class PixelDrawer {
    
    final static public int COPY_DIRECT                 = 0;
    final static public int COPY_FLIP_HORIZONTALLY      = 1;
    final static public int COPY_FLIP_VERTICALLY        = 2;
    
    static public void paintPixel(int x, int y, int rgb, int alpha, 
                                  int[] dst, int dstW, int dstH) {
        if (alpha == 0) {
            return;
        }
        if (alpha == 255) {
            setPixel(x, y, rgb, dst, dstW, dstH);
        } else {
            mixPixel(x, y, rgb, alpha, dst, dstW, dstH);
        }
    }
    
    static public void setPixel(int x, int y, int rgb, 
                                int[] dst, int dstW, int dstH) {
        if (x < 0 || y < 0 || x >= dstW || y >= dstH) {
            return;
        }
        dst[x + y * dstW] = rgb;
    }
    
    static public void mixPixel(int x, int y, int rgb, int alpha,
                                int[] dst, int dstW, int dstH) {
        if (x < 0 || y < 0 || x >= dstW || y >= dstH) {
            return;
        }
        int p = x + y * dstW;
        int rgb0 = dst[p];
        if (rgb == rgb0) {
            return;
        }
        int r0 = rgb0 & RGB.RED;
        int g0 = rgb0 & RGB.GRN;
        int b0 = rgb0 & RGB.BLU;
        int a = alpha + 1;
        dst[p] = (r0 + ((a * ((rgb & RGB.RED) - r0)) >> 8)) & RGB.RED | 
                 (g0 + ((a * ((rgb & RGB.GRN) - g0)) >> 8)) & RGB.GRN | 
                 (b0 + ((a * ((rgb & RGB.BLU) - b0)) >> 8)) & RGB.BLU;
    }
    
    /*
     * An optimized version of mixPixel() for black.
     */
    static public void mixBlack(int x, int y, int alpha,
                                int[] dst, int dstW, int dstH) {
        if (x < 0 || y < 0 || x >= dstW || y >= dstH) {
            return;
        }
        int p = x + y * dstW;
        int rgb0 = dst[p];
        if (rgb0 == 0x000000) {
            return;
        }
        int a = (256 - alpha);
        dst[p] = ((a * ((rgb0 & RGB.RED) >> 16)) << 8) & RGB.RED | 
                  (a * ((rgb0 & RGB.GRN) >>  8))       & RGB.GRN | 
                 ((a *  (rgb0 & RGB.BLU)       ) >> 8) & RGB.BLU;
    }
    
    static public void bitshiftFade(int x, int y, int nBits,
                                    int[] dst, int dstW, int dstH) {
        if (x < 0 || y < 0 || x >= dstW || y >= dstH) {
            return;
        }
        int p = x + y * dstW;
        int rgb0 = dst[p];
        if (rgb0 == 0x000000) {
            return;
        }
        dst[p] = ((rgb0 & RGB.RED) >> nBits) & RGB.RED | 
                 ((rgb0 & RGB.GRN) >> nBits) & RGB.GRN | 
                 ((rgb0 & RGB.BLU) >> nBits) & RGB.BLU;
    }
    
    static public void paintMask(int[] mask, int maskW, int x, int y, int rgb,
                                 int[] dst, int dstW, int dstH) {
        paintMask(mask, maskW, x, y, rgb, 255, dst, dstW, dstH);
    }
    
    static public void paintMask(int[] mask, int maskW, 
                                 int x, int y, int rgb, int alpha,
                                 int[] dst, int dstW, int dstH) {
        int maskH = mask.length / maskW;
        int p = x + y * dstW;
        int k = 0;
        for (int j = 0; j < maskH; j++) {
            for (int i = 0; i < maskW; i++) {
                if (alpha == 255) {
                    paintPixel(x+i, y+j, rgb, mask[k], dst, dstW, dstH);
                } else {
                    paintPixel(x+i, y+j, rgb, (mask[k] * alpha) >> 8,
                               dst, dstW, dstH);
                }
                k++;
            }
        }
    }
    
    static public void copyRect(int srcX, int srcY, int srcW, int srcH, 
                                int[] srcField, int srcFieldW, int srcFieldH,
                                int dstX, int dstY, 
                                int[] dstField, int dstFieldW, int dstFieldH,
                                int copyMode) {
        int i, j, p, q;
        switch (copyMode) {
            case COPY_DIRECT:
                for (j = 0; j < srcH; j++) {
                    System.arraycopy(srcField, srcX + (srcY+j) * srcFieldW, 
                                     dstField, dstX + (dstY+j) * dstFieldW, 
                                     srcW);
                }
                break;
            case COPY_FLIP_HORIZONTALLY:
                for (j = 0; j < srcH; j++) {
                    p = srcX + (srcY+j) * srcFieldW;
                    q = dstX + (dstY+j) * dstFieldW + srcW - 1;
                    for (i = 0; i < srcW; i++) {
                        dstField[q--] = srcField[p++];
                    }
                }
                break;
            case COPY_FLIP_VERTICALLY:
                for (i = 0; i < srcW; i++) {
                    p = srcX + i + srcY * srcFieldW;
                    q = dstX + i + (dstY+srcH-1) * dstFieldW;
                    for (j = 0; j < srcH; j++) {
                        dstField[q] = srcField[p];
                        q -= dstFieldW; p += srcFieldW;
                    }
                }
                break;
        }
    }
    
    static public void paintRect(int srcX, int srcY, int srcW, int srcH, 
                                 int[] srcField, int srcFieldW, int srcFieldH,
                                 int alpha,
                                 int dstX, int dstY, 
                                 int[] dstField, int dstFieldW, int dstFieldH) {
        int p = srcX + srcY * srcFieldW;
        int n = srcFieldW * srcFieldH;
        for (int j = dstY; j < dstY + srcH; j++) {
            for (int i = dstX; i < dstX + srcW; i++) {
                if (p < 0) {
                    continue;
                } else if (p >= n) {
                    return;
                }
                paintPixel(i, j, srcField[p++], alpha, 
                           dstField, dstFieldW, dstFieldH);
            }
            p += srcFieldW - srcW;
        }
    }
    
    static public void drawWuLine(int x1, int y1, int x2, int y2, int rgb,
                                  int[] dst, int dstW, int dstH) {
        drawWuLine(x1, y1, x2, y2, rgb, 255, dst, dstW, dstH);
    }
    
    static public void drawWuLine(int x1, int y1, int x2, int y2, int rgb,
                                  int alpha,
                                  int[] dst, int dstW, int dstH) {
        if (alpha == 0) {
            return;
        }
        long x, y, inc;
        int dx = x2 - x1;
        int dy = y2 - y1;
        if (dx == 0 && dy == 0) {
            return;
        }
        if (Math.abs(dx) > Math.abs(dy)) {
            if (dx < 0) {
                dx = -dx;
                dy = -dy;
                int save;
                save = x1; x1 = x2; x2 = save;
                save = y1; y1 = y2; y2 = save;
            }
            x = x1 << 16;
            y = y1 << 16;
            inc = (dy << 16) / dx;
            while ((x >> 16) < x2) {
                if (alpha == 255) {
                    mixPixel((int) (x >> 16), (int) (y >> 16),
                             rgb, (int) (~y >> 8) & 0xFF, dst, dstW, dstH);
                    mixPixel((int) (x >> 16), (int) (y >> 16) + 1, 
                             rgb, (int) (y >> 8) & 0xFF, dst, dstW, dstH);
                } else {
                    mixPixel((int) (x >> 16), (int) (y >> 16),
                             rgb, (((int) (~y >> 8) & 0xFF) * alpha) >> 8, 
                             dst, dstW, dstH);
                    mixPixel((int) (x >> 16), (int) (y >> 16) + 1, 
                             rgb, (((int) (y >> 8) & 0xFF) * alpha) >> 8, 
                             dst, dstW, dstH);
                }
                x += (1 << 16);
                y += inc;
            }
        } else {
            if (dy < 0) {
                dx = -dx;
                dy = -dy;
                int save;
                save = x1; x1 = x2; x2 = save;
                save = y1; y1 = y2; y2 = save;
            }
            x = x1 << 16;
            y = y1 << 16;
            inc = (dx << 16) / dy;
            while ((y >> 16) < y2) {
                if (alpha == 255) {
                    mixPixel((int) (x >> 16), (int) (y >> 16), 
                             rgb, (int) (~x >> 8) & 0xFF, dst, dstW, dstH);
                    mixPixel((int) (x >> 16) + 1, (int) (y >> 16), 
                             rgb, (int) (x >> 8) & 0xFF, dst, dstW, dstH);
                } else {
                    mixPixel((int) (x >> 16), (int) (y >> 16), 
                             rgb, (((int) (~x >> 8) & 0xFF) * alpha) >> 8, 
                             dst, dstW, dstH);
                    mixPixel((int) (x >> 16) + 1, (int) (y >> 16), 
                             rgb, (((int) (x >> 8) & 0xFF) * alpha) >> 8, 
                             dst, dstW, dstH);
                }
                x += inc;
                y += (1 << 16);
            }
        }
    }
}

