/*
 * com/ngon/tap/EvolCache.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.app.App;
import com.ngon.io.TextFileRunner;
import java.util.Hashtable;
import java.util.StringTokenizer;

/**
 * class EvolCache
 */
class EvolCache implements Evolver {
    
    final static int MAX_RESENDS = 3;
    final static int RESEND_WAIT_COUNT = 10000 / TapEvol.DT;
    
    protected String cacheUrl;
    protected App app;
    protected TextFileRunner fileRunner;
    protected String[] cache;
    protected int cacheIdx;
    protected Hashtable clipInfo;
    protected int nResends;
    protected int resendWaitCount;
    
    EvolCache(String cacheUrl, App app) {
        this.cacheUrl = cacheUrl;
        this.app = app;
        cache = new String[EvolServlet.N_EVOL];
        fileRunner = new TextFileRunner(app);
        if (Debug.DEBUG_APPLET) {
            fileRunner.setVerbose(true);
        }
    }
    
    public void init() {
        fileRunner.start();
        if (clipInfo == null) {
            getClipInfo();
        }
        fileRunner.setRequest(cacheUrl);
    }
    
    public Hashtable getClipInfo() {
        if (clipInfo == null) {
            fileRunner.setRequest(cacheUrl + "?getClipInfo=1");
            if (Debug.DEBUG_APPLET) {
                System.err.println("clipInfo == null, filling...");
            }
            
            // Wait for response.
            String[] response = null;
            int tries = 0;
            while (response == null) {
                try {
                    Thread.sleep(100);
                } catch (InterruptedException e) {}
                if (Debug.DEBUG_APPLET) {
                    System.err.println("getInfoClip tries=" + tries);
                }
                tries++;
                response = fileRunner.getLines();
                if (tries > 300) {
                    return null;
                }
            }
            
            // Parse response into hashtable.
            clipInfo = new Hashtable(response.length * 2);
            for (int i = 0; i < response.length; i++) {
                ClipInfo ci = 
                    ClipInfo.parseEncodedString(response[i],
                                                EvolServlet.FIELD_DELIM);
                if (ci != null) {
                    clipInfo.put(new Integer(ci.clipID), ci);
                }
            }
            fileRunner.clear();
        }
        
        return clipInfo;
    }
    
    public boolean evol(Taps taps, int maxTaps, ClipSegFactory csFactory) {
        // Refill cache if necessary.
        if (cacheIdx == 0) {
            String[] response = fileRunner.getLines();
            if (response == null) {
                // If in the middle of resend, return immediately.
                if (nResends > 0 && resendWaitCount++ < RESEND_WAIT_COUNT) {
                    if (Debug.DEBUG_APPLET) {
                        System.err.println(
                            "resendWaitCount = " + resendWaitCount);
                    }
                    return false;
                }
                // Make sure we have fileRunner active.
                if (fileRunner.getRequest() == null) {
                    if (++nResends > MAX_RESENDS) {
                        return false;
                    }
                    if (Debug.DEBUG_APPLET) {
                        System.err.println(
                            "EvolCache.evol(): fileRunner lost src! Retry " + 
                            nResends);
                    }
                    fileRunner.resendRequest();
                    resendWaitCount = 0;
                }
                if (Debug.DEBUG_APPLET) {
                    System.out.println("evol() returning false");
                }
                return false;
            }
            nResends = 0;
            
            // Copy response to local cache.
            for (int i = 0; i < EvolServlet.N_EVOL; i++) {
                cache[i] = new String(response[i]);
            }
            fileRunner.clear();
            
            // Start fileRunner with next query.
            fileRunner.setRequest(
                cacheUrl + "?max=" + maxTaps + 
                "&ref=" + response[EvolServlet.N_EVOL]);
        }
        
        // Parse current evol line and take action on current taps.
        int tapIdx = 0;
        StringTokenizer stTaps = 
            new StringTokenizer(cache[cacheIdx], EvolServlet.TAP_DELIM);
        while (stTaps.hasMoreTokens()) {
            // Token is of form <action>[|<clus>|<blk0>|...|<blkN>]
            StringTokenizer stTap = 
                new StringTokenizer(stTaps.nextToken(), EvolServlet.ELEM_DELIM);
            
            // First token in line is action code (see TapEvent).
            // Actions may be:
            //   + Add new tap
            //   + Drop tap
            //   + Change existing tap
            String action = stTap.nextToken();
            // Don't let an error bring us down.
            if (tapIdx >= taps.size()) {
                action = TapEvent.EVENT_ADD;
            }
            if (action.compareTo(TapEvent.EVENT_CHANGE) == 0) {
                Tap tap = (Tap) taps.elementAt(tapIdx++);
                if (tap.clus.clip != null) {
                    // Advance to next frame in clip, 
                    // and set the new cluster data.
                    if (tap.clus.clip.nextFrame()) {
                        tap.nFailedAdvances = 0;
                        if (! stTap.hasMoreTokens()) {
                            // Shouldn't happen.
                            continue;
                        }
                        Clus next = 
                            Clus.parseEncodedString(stTap.nextToken(),
                                                    EvolServlet.FIELD_DELIM);
                        if (next != null) {
                            tap.clus.copyFields(next);
                            tap.clus.blocks = getBlocks(stTap);
                        } else {
                            // Really shouldn't get here,
                            // possibly some data / parse error.
                            taps.killTap(tapIdx-1);
                            if (Debug.DEBUG_APPLET) {
                                System.err.println(
                                    "EvolCache.evol() : next clus is null");
                            }
                        }
                    } else {
                        tap.nFailedAdvances++;
                        if (Debug.DEBUG_APPLET) {
                            System.err.println(
                                "Failed to advance frame for clip " +
                                tap.clus.clipID + "." + tap.clus.frameNr + 
                                " fails=" + tap.nFailedAdvances);
                        }
                        if (tap.nFailedAdvances > NULL_FRAME_RETRIES) {
                            // Don't kill entire tap, just nullify clip.
                            taps.killClip(tapIdx-1);
                        }
                    }
                }
            } else if (action.compareTo(TapEvent.EVENT_ADD) == 0) {
                if (! stTap.hasMoreTokens()) {
                    // Shouldn't happen.
                    continue;
                }
                Clus next = 
                    Clus.parseEncodedString(stTap.nextToken(),
                                            EvolServlet.FIELD_DELIM);
                next.clip = csFactory.create(next.clipID, next.frameNr);
                if (next.clip == null) {
                    continue;
                }
                next.clip.start();
                next.blocks = getBlocks(stTap);
                if (tapIdx <= taps.size()) {
                    taps.insertElementAt(new Tap(next), tapIdx);
                } else {
                    taps.addElement(new Tap(next));
                }
            } else if (action.compareTo(TapEvent.EVENT_DROP) == 0) {
                taps.killTap(tapIdx);
                continue;
            }
        }
        
        if (taps.size() > maxTaps) {
            taps.killTap(0);
        }
        cacheIdx = (cacheIdx+1) % EvolServlet.N_EVOL;
        return true;
    }
    
    public void kill() {
        fileRunner.kill();
    }
    
    protected Block[] getBlocks(StringTokenizer stTap) {
        Block[] blocks = new Block[stTap.countTokens()];
        int i = 0;
        while (stTap.hasMoreTokens()) {
            blocks[i++] = Block.parseEncodedString(stTap.nextToken(),
                                                   EvolServlet.FIELD_DELIM);
        }
        return blocks;
    }
}
