/*
 * com/ngon/tap/EvolServlet.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import java.io.IOException;
import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;

public class EvolServlet extends HttpServlet {
    
    final static int N_EVOL = 10;
    
    final static String TAP_DELIM       = "*";
    final static String ELEM_DELIM      = "|";
    final static String FIELD_DELIM     = ",";
    
    protected DataSource dataSrc = null;
    protected int[] initialClusters = null;
    protected Hashtable clipInfo = null;
    
    /**
     * @param stub Arguments passed by the Servlet loader
     * @return void
     */
    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        
        // JDBC connection pool.
        try {
            Context initCtx = new InitialContext();
            Context envCtx = (Context) initCtx.lookup("java:comp/env");
            dataSrc = (DataSource) envCtx.lookup("jdbc/c3cuts");
            
            // Cache initialClusters and clipInfo.
            Connection dbc = null;
            try {
                dbc = dataSrc.getConnection();
                initialClusters = Evol.getInitialClusters(dbc);
                clipInfo = Evol.getClipInfo(dbc);
                dbc.close();
                dbc = null;
            } catch (SQLException e) {
                e.printStackTrace();
            } finally {
                if (dbc != null) {
                    try { dbc.close(); } catch (SQLException e) {}
                    dbc = null;
                }
            }
        } catch (NamingException e) {
            e.printStackTrace();
        }
    }

    /**
     * Main service routine - do a register.
     * @param request The HttpServletRequest.
     * @param response The HttpServletResponse.
     * @throws ServletException
     * @throws IOException
     */
    public void service(HttpServletRequest request, 
                        HttpServletResponse response) throws ServletException,
                                                             IOException {
        
        // Prepare output stream.
        response.setContentType("text/plain");
        PrintWriter writer = response.getWriter();
        
        // If request is "getClipInfo", return that and exit.
        if (request.getParameter("getClipInfo") != null) {
            if (clipInfo == null) {
                throw new ServletException("EvolServlet: clipInfo is null");
            }
            for (Enumeration e = clipInfo.elements(); e.hasMoreElements(); ) {
                ClipInfo ci = (ClipInfo) e.nextElement();
                StringBuffer sb = 
                    ci.toEncodedStringBuffer(EvolServlet.FIELD_DELIM);
                writer.println(sb);
            }
            writer.close();
            return;
        }
        
        // Read HTTP params.
        String referenceTapData = "";
        int maxTaps = 1;
        String s;
        s = request.getParameter("max");
        if (s != null) {
            maxTaps = Integer.parseInt(s);
        }
        s = request.getParameter("ref");
        if (s != null) {
            referenceTapData = s;
        }
        
        // Parse referenceTapData to recreate taps.
        Taps taps = new Taps();
        StringTokenizer stTap = 
            new StringTokenizer(referenceTapData, ELEM_DELIM);
        while (stTap.hasMoreTokens()) {
            Clus c = Clus.parseEncodedString(stTap.nextToken(), FIELD_DELIM);
            if (c != null) {
                taps.addElement(new Tap(c));
            }
        }
        
        // Init evol and db connection from pool.
        Connection dbc = null;
        EvolEvent evol = null;
        try {
            dbc = dataSrc.getConnection();
            evol = new EvolEvent(dbc);
            if (dbc == null || evol == null) {
                throw new SQLException("EvolServlet.service(): BOOM");
            }
            evol.init();
            evol.setInitialClusters(initialClusters);
            evol.setClipInfo(clipInfo);
            
            // Run evol N_EVOL times to build up cache.
            String[] lines = new String[N_EVOL];
            for (int i = 0; i < N_EVOL; i++) {
                evol.evol(taps, maxTaps, null);
                Taps events = evol.getEvents();
                
                // Write out taps as string.
                StringBuffer sb = new StringBuffer();
                for (Enumeration e = events.elements(); e.hasMoreElements(); ) {
                    TapEvent evt = (TapEvent) e.nextElement();
                    Clus c = evt.clus;
                    if (sb.length() > 0) {
                        sb.append(TAP_DELIM);
                    }
                    // Action code.
                    sb.append(evt.actionCode);
                    // Cluster data.
                    if (c == null) {
                        continue;
                    }
                    sb.append(ELEM_DELIM).
                        append(c.toEncodedStringBuffer(FIELD_DELIM));
                    // Blocks.
                    if (c.blocks == null) {
                        c.blocks = evol.getBlocks(c);
                    }
                    if (c.blocks != null) {
                        try {
                            for (int j = 0; j < c.blocks.length; j++) {
                                Block b = c.blocks[j];
                                if (b == null) {
                                    continue;
                                }
                                sb.append(ELEM_DELIM).
                                    append(
                                        b.toEncodedStringBuffer(FIELD_DELIM));
                            }
                        } catch (Exception ex) {
                            continue;
                        }
                    }
                }
                lines[i] = new String(sb);
            }
            
            // Last line is next request 'ref' param.
            StringBuffer nextRef = new StringBuffer();
            for (Enumeration e = taps.elements(); e.hasMoreElements(); ) {
                TapEvent tap = (TapEvent) e.nextElement();
                Clus c = tap.clus;
                if (nextRef.length() > 0) {
                    nextRef.append(ELEM_DELIM);
                }
                nextRef.append(c.toEncodedStringBuffer(FIELD_DELIM));
            }
            
            // Print.
            for (int i = 0; i < lines.length; i++) {
                writer.println(lines[i]);
            }
            writer.println(nextRef);
            
            // Return to connection pool.
            evol.releaseDB();
            evol.kill();
            dbc.close();
            dbc = null;
        } catch (SQLException e) {
            e.printStackTrace();
        } finally {
            if (writer != null) {
                writer.close();
                writer = null;
            }
            if (dbc != null) {
                try { dbc.close(); } catch (SQLException e) {}
                dbc = null;
            }
            if (taps != null) {
                taps.removeAllElements();
                taps = null;
            }
        }
    }
}

